;******************************************************************************
;*      INTERRUPT[X] - procedures for calling interrupt service routines      *
;*           Replacement for QB/VBDOS's INTERRUPT[X] procedures               *
;*                                                                            *
;*                       (c) Hans Lunsing -  8/1994                           *
;*                                                                            *
;* For QuickBasic 4.5, Basic PDS (QB eXtended) 7.x en Visual Basic for DOS 1.0*
;******************************************************************************

.model          medium, basic

;-----------------------------------------------------------------------------
; Interrupt and InterruptX - BASIC interrupt calling interface
;
; Purpose:
;       To allow a BASIC Compiler program to perform any software
;       interrupt.  The interrupt is executed with the registers
;       set to values specified in a register variable.  The post-
;       interrupt values of the registers are then stored in
;       another register variable.
;
;       CALL Interrupt[X] (IntNo  AS INTEGER,
;                          InReg AS RegType[X],
;                          OutReg AS RegType[X])
;
; Inputs:
;       IntNo  = interrupt number (range 0 to 255) to execute
;       InReg and OutReg are register variables of type RegType[X]
;       defined as follows;
;
;       TYPE RegType
;           ax    AS INTEGER
;           bx    AS INTEGER
;           cx    AS INTEGER
;           dx    AS INTEGER
;           bp    AS INTEGER
;           si    AS INTEGER
;           di    AS INTEGER
;           flags AS INTEGER
;       END TYPE
;
;       TYPE RegTypeX
;           ax    AS INTEGER
;           bx    AS INTEGER
;           cx    AS INTEGER
;           dx    AS INTEGER
;           bp    AS INTEGER
;           si    AS INTEGER
;           di    AS INTEGER
;           flags AS INTEGER
;           ds    AS INTEGER
;           es    AS INTEGER
;       END TYPE
;
;       To leave BP, DS and ES unchanged input them as -1.
;
; Outputs:
;       If no error:
;               IntNo  = unchanged (range 0 to 255)
;               OutReg: This array will be set to the post-interrupt
;                       register values.  It has the same structure
;                       as InReg.
;       If error (bad int number):
;               IntNo  = -1
;               OutReg unchanged.  INT call is not performed.
;               Error occurs if first argument not in range 0 to 255 (2^8-1)
;
; Modifies:
;       All, except BP, DS, SI, DI and flags.
;       Also, possible side effects of INT call.
;
; Exceptions:
;       INT 24H call may result from some INT 21H MsDOS calls.
;       Therefore BP cannot be changed when executing a DOS interrupt
;       (in the range 20h - 2Fh). In fact it is never necessary to
;       use BP in a DOS interrupt, so this is not really a limitation.
;-----------------------------------------------------------------------------

; Frame structure definition

IntNo           EQU  <[BP+0CH]>         ;pointer to first of three arguments
InReg           EQU  <[BP+0AH]>         ;pointer to second of three arguments
OutReg          EQU  <[BP+08H]>         ;pointer to third of three arguments
;PM             EQU  <[BP+04H]>         ;far return addres
;PM             EQU  <[BP+02H]>         ;near return addres for AllCOmmon
;PM             EQU  <[BP    ]>         ;saved base pointer
Status          EQU  <[BP-02H]>         ;lo byte is status byte
DataSeg         EQU  <[BP-04H]>         ;callers data segment

; Register structure definition

Int_AX          EQU <[SI    ]>          ;Int AX register value
Int_BX          EQU <[SI+02H]>          ;Int BX register value
Int_CX          EQU <[SI+04H]>          ;Int CX register value
Int_DX          EQU <[SI+06H]>          ;Int DX register value
Int_BP          EQU <[SI+08H]>          ;Int BP register value
Int_SI          EQU <[SI+0AH]>          ;Int SI register value
Int_DI          EQU <[SI+0CH]>          ;Int DI register value
Int_Flgs        EQU <[SI+0EH]>          ;Int flags register value
Int_DS          EQU <[SI+10H]>          ;Int DS register value
Int_ES          EQU <[SI+12H]>          ;Int ES register value

;-----------------------------------------------------------------------------
.CODE

Interrupt       PROC    FAR

                MOV     AL, 1           ;eight regs used (not DS or ES)
                CALL    NEAR PTR InterruptCommon
                RETF    6

Interrupt       ENDP


InterruptX      PROC    FAR

                XOR     AL, AL          ;ten regs used (including DS and ES)
                CALL    NEAR PTR InterruptCommon
                RETF    6

InterruptX      ENDP


InterruptCommon:
                PUSH    BP              ;save callers stack frame pointer
                MOV     BP, SP          ;establish program frame reference

                MOV     BX, IntNo       ;Get interruptnumber
                MOV     DX, [BX]
                OR      DH, DH          ;test if in range
                JZ      @F              ;if OK then continue

                ; If error, then set IntNo  to -1 to report error and exit

                MOV     WORD PTR [BX], -1 ;set interrupt number to -1 for
                                        ;error
                JMP     Int_Exit        ;and exit

@@:             MOV     CS:ExecInt[1], DL  ;store interruptnumber

                CMP     DL,25H          ;Interrupt 25 request?
                JZ      Int2526         ;Jump if so
                CMP     DL,26H          ;Interrupt 26 request?
                JNZ     IntOther        ;Jump if other, "normal" interrupt
Int2526:        MOV     BX, 4444h       ;Int 25h or 26h leave two words of
                                        ;flags on the stack, which must be
                                        ;cleaned after the interrupt.
                                        ;44h is opcode for INC SP
                JMP     SHORT @F
IntOther:       MOV     BX, 9090h       ;90h is opcode for NOP
@@:             MOV     WORD PTR CS:ExecInt[2], BX

                ; Save status byte on stack:
                ; AL bit 0 = (if 8 registers then 1 else 0)

                PUSH    AX              ;save AL

                ; Save Basic's DS, ES, SI, DI and processor flags on the stack

                PUSH    DS              ;DS saved as DataSeg
                PUSH    ES
                PUSH    SI
                PUSH    DI
                PUSHF

                ; Move the InReg values to their actual registers.

                MOV     SI, InReg       ;Point SI to InReg structure

                ; For BP, leave in base pointer when executing MsDOS, or
                ; when executing Interrupt(X) with BP element value of -1.

                CMP     DL, 20h         ;test if IntNo (yet in DL) is in
                JL      Int_TestBP      ;MsDOS interrupt range 20h to 2Fh
                CMP     DL, 30h
                JL      @F              ;if so then no change of BP

Int_TestBP:     CMP     WORD PTR Int_BP, -1 ;test if default BP to be used
                JE      @F              ;if so, then leave it unchanged
                MOV     BP, Int_BP      ;set up input BP value

@@:             MOV     AX, Int_Flgs    ;get input flag register value
                AND     AX, 0000111111010101B   ;mask out undefined 8086 flags
                PUSH    AX              ;push masked flag register value

                MOV     AX, Int_AX      ;set up input AX value
                MOV     BX, Int_BX      ;set up input BX value
                MOV     CX, Int_CX      ;set up input CX value
                MOV     DX, Int_DX      ;set up input DX value

                MOV     DI, Int_DI      ;set up input DI value
                PUSH    Int_SI          ;save Int_SI on stack

                ; For DS and ES, leave in Basic's data segment values
                ; when executing Interrupt, or when executing InterruptX
                ; or MsDOS with element values of -1.

                TEST    BYTE PTR Status, 1    ;test if executing Interrupt
                JNZ     @F              ;if so, then use both default values

                CMP     WORD PTR Int_ES, -1   ;test if default ES to be used
                JE      Int_SetupDS     ;if so, then leave it unchanged
                MOV     ES, Int_ES      ;set up input ES value

Int_SetupDS:    CMP     WORD PTR Int_DS, -1   ;test if default DS to be used
                JE      @F              ;if so, then leave it unchanged
                MOV     DS, Int_DS      ;set up input DS value
                                        ;must be last move involving DS:SI

@@:             POP     SI              ;set up input SI value
                POPF                    ;set up input flag register value

                ; With all registers set according to the inreg
                ; structure, execute the interrupt.

ExecInt LABEL BYTE
                DB      0CDh, ?         ;opcode for: INT ?
                DW      ?               ;opcode for 2x INC SP (INT 25/26)
                                        ;or 2x NOP (Other)

                ; First save the BP and flags values from the INT call.
                ; Then get the old value of BP to reference the frame by
                ; adjusting the value of the stack pointer for items
                ; pushed on the stack since establishing the frame: the
                ; status word, callers DS, ES, SI, DI and flags, and the
                ; BP value from the INT call.

                PUSH    BP
                MOV     BP,SP           ;temporary frame
                PUSHF                   ;save flags on the stack
                ADD     BP,14           ;adjust for items pushed on stack

                ; Save DS and SI values from the INT call and point
                ; DS:SI to OutReg structure

@@:             PUSH    DS
                PUSH    SI

                MOV     DS, DataSeg     ;Retrieve Basic's data segment
                MOV     SI, OutReg      ;Point SI to OutReg structure

                ; Put post-INT value of all registers into the OutReg structure

                MOV     Int_AX, AX      ;put in post-INT AX value
                MOV     Int_BX, BX      ;put in post-INT BX value
                MOV     Int_CX, CX      ;put in post-INT CX value
                MOV     Int_DX, DX      ;put in post-INT DX value

                POP     Int_SI          ;put in post-INT SI value
                MOV     Int_DI, DI      ;put in post-INT DI value

                POP     DX              ;pop post_INT DS-value

                MOV     AL, BYTE PTR Status   ;Status byte in AL
                TEST    AL, 1           ;don't put in ES and DS when calling
                JNZ     @F              ;Interrupt
                MOV     Int_ES, ES      ;put in post-INT ES value
                MOV     Int_DS, DX      ;put in post-INT DS value

@@:             POP     Int_Flgs        ;put in post-INT flag register value
                POP     Int_BP          ;put in post-INT BP value

                ; Restore Basic's processor flags, DI, SI, ES and DS

@@:             POPF
                POP     DI
                POP     SI
                POP     ES
                POP     DS              ;DS was retrieved already
                POP     AX

Int_Exit:       POP     BP              ;restore caller's stack frame
                RETN

END
